#| -*-Scheme-*-

Copyright (C) 2007, 2008, 2009, 2010, 2011, 2012, 2013, 2014, 2015,
    2016, 2017, 2018, 2019 Matthew Birkholz

This file is part of a Cairo plugin for MIT/GNU Scheme Pucked.

This plugin is free software; you can redistribute it and/or modify it
under the terms of the GNU General Public License as published by the
Free Software Foundation; either version 2 of the License, or (at your
option) any later version.

This plugin is distributed in the hope that it will be useful, but
WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
General Public License for more details.

You should have received a copy of the GNU General Public License
along with this plugin; if not, write to the Free Software Foundation,
Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301, USA.

|#

;;;; Cairo Objects
;;; package: (gtk cairo)

(C-include "cairo")

(define-integrable (color-red o) (flo:vector-ref o 0))
(define-integrable (color-green o) (flo:vector-ref o 1))
(define-integrable (color-blue o) (flo:vector-ref o 2))
(define-integrable (color-alpha o) (flo:vector-ref o 3))

(define (rgba-color red green blue alpha)
  (let ((color (flo:vector-cons 4)))
    (flo:vector-set! color 0 red)
    (flo:vector-set! color 1 green)
    (flo:vector-set! color 2 blue)
    (flo:vector-set! color 3 alpha)
    color))

(define white (rgba-color 1. 1. 1. 1.))
(define black (rgba-color 0. 0. 0. 1.))

(define (guarantee-color object operator)
  (if (not (and (flo:flonum? object) (fix:= 4 (flo:vector-length object))))
      (error:wrong-type-argument object "an rgba color" operator)))

(define (cairo-image-surface-create width height)
  (let ((surface (make-alien '|cairo_surface_t|))
	(copy (make-alien '|cairo_surface_t|)))
    (add-cairo-cleanup surface (make-cairo-surface-cleanup copy))
    (C-call "cairo_image_surface_create" copy
	    (C-enum "CAIRO_FORMAT_RGB24") width height)
    (copy-alien-address! surface copy)
    (check-cairo-surface-status surface)
    surface))

(define (make-cairo-surface-cleanup alien)
  (named-lambda (cairo-surface-cleanup)
    (assert-without-interruption 'cairo-surface-cleanup)
    (if (not (alien-null? alien))
	(begin
	  (C-call "cairo_surface_destroy" alien)
	  (alien-null! alien)))))

(define (cairo-surface-destroy surface)
  (check-cairo-surface-status surface)
  (without-interruption
   (lambda ()
     (if (not (alien-null? surface))
	 (execute-cairo-cleanup surface)))))

(define (check-cairo-surface-status surface)
  (let ((status (C-call "cairo_surface_status" surface)))
    (if (not (= status (C-enum "CAIRO_STATUS_SUCCESS")))
	(let ((msg (C-call "cairo_status_to_string"
			   (make-alien '(* (const char)))
			   status)))
	  (error (c-peek-cstring msg) surface)))))

(define (guarantee-cairo-surface object operator)
  (if (and (alien? object) (eq? (alien/ctype object) '|cairo_surface_t|))
      object
      (error:wrong-type-argument object "a cairo_surface_t alien" operator)))

(define (cairo-surface-write-to-png surface filename)
  (guarantee-cairo-surface surface 'cairo-surface-write-to-png)
  (guarantee string? filename 'cairo-surface-write-to-png)
  (C-call "cairo_surface_write_to_png" surface (string->utf8 filename)))

(define (cairo-surface-flush surface)
  (guarantee-cairo-surface surface 'cairo-surface-flush)
  (C-call "cairo_surface_flush" surface))

(define (cairo-pattern-create-linear x1 y1 x2 y2)
  (let ((pattern (make-alien '|cairo_pattern_t|))
	(copy (make-alien '|cairo_pattern_t|)))
    (add-cairo-cleanup pattern (make-cairo-pattern-cleanup copy))
    (C-call "cairo_pattern_create_linear" copy
	    (->flonum x1) (->flonum y1) (->flonum x2) (->flonum y2))
    (copy-alien-address! pattern copy)
    (check-cairo-pattern-status pattern)
    pattern))

(define (cairo-pattern-create-radial cx0 cy0 radius0 cx1 cy1 radius1)
  (let ((pattern (make-alien '|cairo_pattern_t|))
	(copy (make-alien '|cairo_pattern_t|)))
    (add-cairo-cleanup pattern (make-cairo-pattern-cleanup copy))
    (C-call "cairo_pattern_create_radial" copy
	    (->flonum cx0) (->flonum cy0) (->flonum radius0)
	    (->flonum cx1) (->flonum cy1) (->flonum radius1))
    (copy-alien-address! pattern copy)
    (check-cairo-pattern-status pattern)
    pattern))

(define (make-cairo-pattern-cleanup alien)
  (named-lambda (cairo-pattern-cleanup)
    (assert-without-interruption 'cairo-pattern-cleanup)
    (if (not (alien-null? alien))
	(begin
	  (C-call "cairo_pattern_destroy" alien)
	  (alien-null! alien)))))

(define (cairo-pattern-destroy pattern)
  (check-cairo-pattern-status pattern)
  (without-interruption
   (lambda ()
     (if (not (alien-null? pattern))
	 (execute-cairo-cleanup pattern)))))

(define (check-cairo-pattern-status pattern)
  (let ((status (C-call "cairo_pattern_status" pattern)))
    (if (not (= status (C-enum "CAIRO_STATUS_SUCCESS")))
	(let ((msg (C-call "cairo_status_to_string"
			   (make-alien '(* (const char)))
			   status)))
	  (error (c-peek-cstring msg) pattern)))))

(define (guarantee-cairo-pattern object operator)
  (if (and (alien? object) (eq? (alien/ctype object) '|cairo_pattern_t|))
      object
      (error:wrong-type-argument object "a cairo_pattern_t alien" operator)))

(define (cairo-pattern-add-color-stop pattern offset color)
  (guarantee-cairo-pattern pattern 'cairo-pattern-add-color-stop)
  (guarantee-color color 'cairo-pattern-add-color-stop)
  (C-call "cairo_pattern_add_color_stop_rgba"
	  pattern (->flonum offset)
	  (color-red color) (color-green color) (color-blue color)
	  (color-alpha color)))

(define (cairo-create surface)
  (guarantee-cairo-surface surface 'cairo-create)
  (let ((cairo (make-alien '|cairo_t|))
	(copy (make-alien '|cairo_t|)))
    (add-cairo-cleanup cairo (make-cairo-cleanup copy))
    (C-call "cairo_create" copy surface)
    (copy-alien-address! cairo copy)
    (check-cairo-status cairo)
    cairo))

(define (make-cairo-cleanup alien)
  (named-lambda (cairo-cleanup)
    (assert-without-interruption 'cairo-cleanup)
    (if (not (alien-null? alien))
	(begin
	  (C-call "cairo_destroy" alien)
	  (alien-null! alien)))))

(define (cairo-destroy cairo)
  (check-cairo-status cairo)
  (without-interruption
   (lambda ()
     (execute-cairo-cleanup cairo))))

(define (check-cairo-status cairo)
  (let ((status (C-call "cairo_status" cairo)))
    (if (not (= status (C-enum "CAIRO_STATUS_SUCCESS")))
	(let ((msg (C-call "cairo_status_to_string"
			   (make-alien '(* (const char)))
			   status)))
	  (error (c-peek-cstring msg) cairo)))))

(define (guarantee-cairo object operator)
  (if (and (alien? object) (eq? (alien/ctype object) '|cairo_t|))
      object
      (error:wrong-type-argument object "a cairo_t alien" operator)))

(define (cairo-save cairo)
  (guarantee-cairo cairo 'cairo-save)
  (C-call "cairo_save" cairo))

(define (cairo-restore cairo)
  (guarantee-cairo cairo 'cairo-save)
  (C-call "cairo_restore" cairo))

(define (cairo-translate cairo dx dy)
  (guarantee-cairo cairo 'cairo-translate)
  (C-call "cairo_translate" cairo (->flonum dx) (->flonum dy)))

(define (cairo-scale cairo sx sy)
  (guarantee-cairo cairo 'cairo-scale)
  (C-call "cairo_scale" cairo (->flonum sx) (->flonum sy)))

(define (cairo-set-source-color cairo color)
  (guarantee-cairo cairo 'cairo-set-source-color)
  (guarantee-color color 'cairo-set-source-color)
  (C-call "cairo_set_source_rgba" cairo
	  (color-red color) (color-green color) (color-blue color)
	  (color-alpha color)))

(define (cairo-set-source cairo pattern)
  (guarantee-cairo cairo 'cairo-set-source)
  (guarantee-cairo-pattern pattern 'cairo-set-source)
  (C-call "cairo_set_source" cairo pattern))

(define (cairo-set-source-surface cairo surface x y)
  (guarantee-cairo cairo 'cairo-set-source-surface)
  (guarantee-cairo-surface surface 'cairo-set-source-surface)
  (C-call "cairo_set_source_surface" cairo surface (->flonum x) (->flonum y)))

(define (cairo-reset-clip cairo)
  (guarantee-cairo cairo 'cairo-reset-clip)
  (C-call "cairo_reset_clip" cairo))

(define (cairo-clip cairo)
  (guarantee-cairo cairo 'cairo-clip)
  (C-call "cairo_clip" cairo))

(define-integrable (cairo-clip-extents cairo receiver)
  (let ((doubles (malloc (fix:* 4 (C-sizeof "double")) 'double)))
    (let ((y1 (C-array-loc doubles "double" 1))
	  (x2 (C-array-loc doubles "double" 2))
	  (y2 (C-array-loc doubles "double" 3)))
      (C-call "cairo_clip_extents" cairo doubles y1 x2 y2)
      (let ((x1. (C-> doubles "double")) (y1. (C-> y1 "double"))
	    (x2. (C-> x2 "double")) (y2. (C-> y2 "double")))
	(free doubles)
	(receiver x1. y1. x2. y2.)))))

(define (cairo-move-to cairo x y)
  (guarantee-cairo cairo 'cairo-move-to)
  (let ((x (->flonum x))
	(y (->flonum y)))
    (C-call "cairo_move_to" cairo x y)))

(define (cairo-close-path cairo)
  (guarantee-cairo cairo 'cairo-close-path)
  (C-call "cairo_close_path" cairo))

(define (cairo-new-sub-path cairo)
  (guarantee-cairo cairo 'cairo-new-sub-path)
  (C-call "cairo_new_sub_path" cairo))

(define (cairo-set-operator cairo operator)
  (guarantee-cairo cairo 'cairo-set-operator)
  (C-call "cairo_set_operator" cairo
	  (case operator
	    ((clear)		(C-enum "CAIRO_OPERATOR_CLEAR"))
	    ((source)		(C-enum "CAIRO_OPERATOR_SOURCE"))
	    ((over)		(C-enum "CAIRO_OPERATOR_OVER"))
	    ((in)		(C-enum "CAIRO_OPERATOR_IN"))
	    ((out)		(C-enum "CAIRO_OPERATOR_OUT"))
	    ((atop)		(C-enum "CAIRO_OPERATOR_ATOP"))
	    ((dest)		(C-enum "CAIRO_OPERATOR_DEST"))
	    ((dest-over)	(C-enum "CAIRO_OPERATOR_DEST_OVER"))
	    ((dest-in)		(C-enum "CAIRO_OPERATOR_DEST_IN"))
	    ((dest-out)		(C-enum "CAIRO_OPERATOR_DEST_OUT"))
	    ((dest-atop)	(C-enum "CAIRO_OPERATOR_DEST_ATOP"))
	    ((xor)		(C-enum "CAIRO_OPERATOR_XOR"))
	    ((add)		(C-enum "CAIRO_OPERATOR_ADD"))
	    ((source)		(C-enum "CAIRO_OPERATOR_SOURCE"))
	    ((saturate)		(C-enum "CAIRO_OPERATOR_SATURATE"))
	    ((multiply)		(C-enum "CAIRO_OPERATOR_MULTIPLY"))
	    ((screen)		(C-enum "CAIRO_OPERATOR_SCREEN"))
	    ((overlay)		(C-enum "CAIRO_OPERATOR_OVERLAY"))
	    ((darken)		(C-enum "CAIRO_OPERATOR_DARKEN"))
	    ((lighten)		(C-enum "CAIRO_OPERATOR_LIGHTEN"))
	    ((color-dodge)	(C-enum "CAIRO_OPERATOR_COLOR_DODGE"))
	    ((color-burn)	(C-enum "CAIRO_OPERATOR_COLOR_BURN"))
	    ((hard-light)	(C-enum "CAIRO_OPERATOR_HARD_LIGHT"))
	    ((soft-light)	(C-enum "CAIRO_OPERATOR_SOFT_LIGHT"))
	    ((difference)	(C-enum "CAIRO_OPERATOR_DIFFERENCE"))
	    ((exclusion)	(C-enum "CAIRO_OPERATOR_EXCLUSION"))
	    ((hsl-hue)		(C-enum "CAIRO_OPERATOR_HSL_HUE"))
	    ((hsl-saturation)	(C-enum "CAIRO_OPERATOR_HSL_SATURATION"))
	    ((hsl-color)	(C-enum "CAIRO_OPERATOR_HSL_COLOR"))
	    ((hsl-luminosity)	(C-enum "CAIRO_OPERATOR_HSL_LUMINOSITY"))
	    (else (error:wrong-type-argument operator "a drawing operator"
					     'cairo-set-operator)))))

(define (cairo-set-line-width cairo width)
  (guarantee-cairo cairo 'cairo-set-line-width)
  (C-call "cairo_set_line_width" cairo (->flonum width)))

(define (cairo-set-dash cairo dashes)
  (guarantee-cairo cairo 'cairo-set-dash)
  (let* ((num (length dashes))
	 (alien (malloc (fix:* num (C-sizeof "double")) 'double))
	 (scan (copy-alien alien)))
    (for-each
      (lambda (len)
	(C->= scan "double" (->flonum len))
	(alien-byte-increment! scan (C-sizeof "double")))
      dashes)
    (C-call "cairo_set_dash" cairo alien num 0)
    (free alien)))

(define (cairo-line-to cairo x y)
  (guarantee-cairo cairo 'cairo-line-to)
  (let ((x (->flonum x))
	(y (->flonum y)))
    (C-call "cairo_line_to" cairo x y)))

(define (cairo-rel-line-to cairo dx dy)
  (guarantee-cairo cairo 'cairo-rel-line-to)
  (let ((dx (->flonum dx))
	(dy (->flonum dy)))
    (C-call "cairo_rel_line_to" cairo dx dy)))

(define (cairo-rectangle cairo x y width height)
  (guarantee-cairo cairo 'cairo-rectangle)
  (C-call "cairo_rectangle" cairo
	  (->flonum x) (->flonum y) (->flonum width) (->flonum height)))

(define (cairo-arc cairo xc yc radius start-angle end-angle)
  (guarantee-cairo cairo 'cairo-arc)
  (C-call "cairo_arc" cairo (->flonum xc) (->flonum yc) (->flonum radius)
	  (->flonum start-angle) (->flonum end-angle)))

(define (cairo-show-pango-layout cairo layout)
  (guarantee-cairo cairo 'cairo-show-pango-layout)
  (guarantee-pango-layout layout 'cairo-show-pango-layout)
  (assert-glib-locked 'cairo-show-pango-layout)
  (C-call "pango_cairo_show_layout" cairo (gobject-alien layout)))

(define (cairo-show-text cairo string)
  (guarantee-cairo cairo 'cairo-show-text)
  (guarantee string? string 'cairo-show-text)
  (C-call "cairo_show_text" cairo (string->utf8 string)))

(define (cairo-paint cairo)
  (guarantee-cairo cairo 'cairo-paint)
  (C-call "cairo_paint" cairo))

(define (cairo-fill cairo)
  (guarantee-cairo cairo 'cairo-fill)
  (C-call "cairo_fill" cairo))

(define (cairo-fill-preserve cairo)
  (guarantee-cairo cairo 'cairo-fill-preserve)
  (C-call "cairo_fill_preserve" cairo))

(define (cairo-stroke cairo)
  (guarantee-cairo cairo 'cairo-stroke)
  (C-call "cairo_stroke" cairo))

(define (cairo-stroke-preserve cairo)
  (guarantee-cairo cairo 'cairo-stroke-preserve)
  (C-call "cairo_stroke_preserve" cairo))

(define (cairo-set-font-matrix cairo matrix)
  (guarantee-cairo cairo 'cairo-set-font-matrix)
  (guarantee-cairo-matrix matrix 'cairo-set-font-matrix)
  #;(C-call "scm_cairo_set_font_matrix" cairo matrix)
  (let ((cairo-matrix (malloc (C-sizeof "cairo_matrix_t") '|cairo_matrix_t|)))
    (C->= cairo-matrix "cairo_matrix_t xx" (flo:vector-ref matrix 0))
    (C->= cairo-matrix "cairo_matrix_t yx" (flo:vector-ref matrix 1))
    (C->= cairo-matrix "cairo_matrix_t x0" (flo:vector-ref matrix 2))
    (C->= cairo-matrix "cairo_matrix_t xy" (flo:vector-ref matrix 3))
    (C->= cairo-matrix "cairo_matrix_t yy" (flo:vector-ref matrix 4))
    (C->= cairo-matrix "cairo_matrix_t y0" (flo:vector-ref matrix 5))
    (C-call "cairo_set_font_matrix" cairo cairo-matrix)
    (free cairo-matrix)))

(define (guarantee-cairo-matrix object operator)
  (if (and (flo:flonum? object)
	   (= (flo:vector-length object) 6))
      object
      (error:wrong-type-argument object "a cairo matrix" operator)))

(define (cairo-matrix xx yx x0
		      xy yy y0)
  (let ((v (flo:vector-cons 6)))
    (flo:vector-set! v 0 xx)
    (flo:vector-set! v 1 yx)
    (flo:vector-set! v 2 x0)
    (flo:vector-set! v 3 xy)
    (flo:vector-set! v 4 yy)
    (flo:vector-set! v 5 y0)
    v))

(define-integrable (xx m) (flo:vector-ref m 0))
(define-integrable (yx m) (flo:vector-ref m 1))
(define-integrable (x0 m) (flo:vector-ref m 2))
(define-integrable (xy m) (flo:vector-ref m 3))
(define-integrable (yy m) (flo:vector-ref m 4))
(define-integrable (y0 m) (flo:vector-ref m 5))

(define-integrable (set-xx! m v) (flo:vector-set! m 0 v))
(define-integrable (set-yx! m v) (flo:vector-set! m 1 v))
(define-integrable (set-x0! m v) (flo:vector-set! m 2 v))
(define-integrable (set-xy! m v) (flo:vector-set! m 3 v))
(define-integrable (set-yy! m v) (flo:vector-set! m 4 v))
(define-integrable (set-y0! m v) (flo:vector-set! m 5 v))

(define (cairo-identity-matrix)
  (cairo-matrix 1. 0. 0.
		0. 1. 0.))

(define (cairo-rotation-matrix radians)
  (let ((c (cos radians))
	(s (sin radians)))
    (cairo-matrix    c  s 0.
		  (- s) c 0.)))

(define (cairo-matrix-scale! matrix factor)
  (guarantee-flonum factor 'cairo-scale-matrix!)
  (set-xx! matrix (flo:* (xx matrix) factor))
  (set-yy! matrix (flo:* (yy matrix) factor)))

(define (cairo-matrix-translate! matrix dx dy)
  (guarantee-flonum dx 'cairo-matrix-translate!)
  (guarantee-flonum dy 'cairo-matrix-translate!)
  (set-x0! matrix (flo:+ dx (x0 matrix)))
  (set-y0! matrix (flo:+ dy (y0 matrix))))

(define (cairo-point x y)
  (let ((v (flo:vector-cons 2)))
    (flo:vector-set! v 0 x)
    (flo:vector-set! v 1 y)
    v))

(define-integrable (x p) (flo:vector-ref p 0))
(define-integrable (y p) (flo:vector-ref p 1))

(define-integrable (set-x! p x) (flo:vector-set! p 0 x))
(define-integrable (set-y! p y) (flo:vector-set! p 1 y))

(define (cairo-transform! point matrix)
  (let ((px (x point))
	(py (y point)))
    (set-x! point (flo:+ (flo:+ (flo:* (xx matrix) px)
				(flo:* (xy matrix) py))
			 (x0 matrix)))
    (set-y! point (flo:+ (flo:+ (flo:* (yx matrix) px)
				(flo:* (yy matrix) py))
			 (y0 matrix)))))

(define (guarantee-flonum object operator)
  (if (not (flo:flonum? object))
      (error:wrong-type-argument object "a flonum" operator)))

(define cairo-cleanups)
(define cairo-cleanups-mutex)

(define (cleanup-cairo!)
  (with-thread-mutex-try-lock cairo-cleanups-mutex
    (lambda ()
      (let loop ((cleanups cairo-cleanups)
		 (prev #f))
	(if (pair? cleanups)
	    (if (weak-pair/car? (car cleanups))
		(loop (cdr cleanups) cleanups)
		(let ((next (cdr cleanups)))
		  ((weak-cdr (car cleanups)))
		  (if prev
		      (set-cdr! prev next)
		      (set! cairo-cleanups next))
		  (loop next prev))))))
    (lambda ()
      unspecific)))

(define (add-cairo-cleanup object thunk)
  (let ((weak (weak-cons object thunk)))
    (with-thread-mutex-lock cairo-cleanups-mutex
      (lambda ()
	(set! cairo-cleanups (cons weak cairo-cleanups))))
    weak))

(define (execute-cairo-cleanup object)
  (assert-without-interruption 'execute-cairo-cleanup)
  (with-thread-mutex-lock cairo-cleanups-mutex
    (lambda ()
      (without-interruption
       (lambda ()
	 (let ((entry (weak-assq object cairo-cleanups)))
	   (if entry
	       (begin
		 ((weak-cdr entry))
		 (set! cairo-cleanups (delq! entry cairo-cleanups)))))))))
  unspecific)

(define (weak-assq obj alist)
  (let loop ((alist alist))
    (if (null? alist)
	#f
	(let* ((entry (car alist))
	       (key (weak-car entry)))
	  (if (eq? obj key)
	      entry
	      (loop (cdr alist)))))))

(define (reset-cairo-package!)
  (set! cairo-cleanups '())
  (set! cairo-cleanups-mutex (make-thread-mutex)))

(define (initialize-package!)
  (reset-cairo-package!)
  (add-event-receiver! event:after-restore reset-cairo-package!)
  (add-gc-daemon! cleanup-cairo!))

(initialize-package!)